/* This program will estimate maximum likelihood branch  */
/* lengths for a fixed species tree given gene trees and */
/* branch lengths under the coalescent process.          */

/* Written by L. Kubatko (2/8/07), copyright 2007        */
/* Version 1.0, August, November 2008, copyright 2008    */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <memory.h>
#include <string.h>
#include <assert.h>
#include <sys/time.h>
#include "main.h"


int ntaxa, nspecies;
int ngtrees, **ppMatrix, *parents, *ppTwoRow[2], *ppTwoRowS[2], *ppTwoRow_temp[2], *ppTwoRowS_temp[2], ***pppTR, ***pppTRS, *filled_ind, *filled_indS, *CoalVec, *CoalVec2, *CoalVecL, *CoalVecR, *recalced, *SpecCount, **SpecVec, **ppSpecConUsed, counteri,burnin, extra_burnin, bound_total_iter, search;
double lik, *TimeVec, *TimeVec_temp, *TimeVecS, *TimeVecS_temp, **ppTimeVec, **ppTimeVecS, *ChangeNodeLength, **ppLengthMat, **ppLengthMatS, **ppSpecCon, **ppLikMat, **ppTipMat, previous_lik, ci;
float beta, mu, theta;
int no_topo_stop, no_to_stop, no_topo_changes, no_accept, no_last_un, bound_last_un, cbound=10, bound_un, cbound_un, done_burnin=0, verbose;

Link Head, currenttree;
int NodeCount=0, NodeCount2=0, use_last=0, last_delete;
int first_annealing_pass=1, num_saved_trees;
naym *taxname=NULL, *speciesname=NULL;

int MemAlloc() {

  int i, j;

  TimeVec = (double*)malloc((2*ntaxa+1)*sizeof(double));
  if (TimeVec==NULL)
    {
      printf("Can't memalloc TimeVec\n");
      return 0;
    }

  TimeVecS = (double*)malloc((2*nspecies+1)*sizeof(double));
  if (TimeVecS==NULL)
    {
      printf("Can't memalloc TimeVec\n");
      return 0;
    }

  TimeVec_temp = (double*)malloc((2*ntaxa+1)*sizeof(double));
  if (TimeVec_temp==NULL)
    {
      printf("Can't memalloc TimeVec_temp\n");
      return 0;
    }

  TimeVecS_temp = (double*)malloc((2*nspecies+1)*sizeof(double));
  if (TimeVecS_temp==NULL)
    {
      printf("Can't memalloc TimeVec_temp\n");
      return 0;
    }

  ppTimeVec = (double**)malloc(ngtrees*sizeof(double*));
  for (i=0; i<ngtrees; i++) {
    ppTimeVec[i] = (double*)malloc((2*ntaxa+1)*sizeof(double));
  }

  ppTimeVecS = (double**)malloc((2*num_saved_trees)*sizeof(double*));
  for (i=0; i<2*num_saved_trees; i++) {
    ppTimeVecS[i] = (double*)malloc((2*nspecies+1)*sizeof(double));
  }


  for (i=0; i<2; i++)
    {
      ppTwoRow[i] = (int*)malloc((2*ntaxa)*sizeof(int));
      if (ppTwoRow[i]==NULL)
	{
	  printf("Can't memalloc ppTwoRow[%d]\n",i);
	  return 0;
	}
    }

  for (i=0; i<2; i++)
    {
      ppTwoRowS[i] = (int*)malloc((2*nspecies)*sizeof(int));
      if (ppTwoRowS[i]==NULL)
        {
          printf("Can't memalloc ppTwoRowS[%d]\n",i);
          return 0;
        }
    }

  ppLikMat = (double**)malloc((ngtrees+1)*sizeof(double*));
  for (i=0; i<(ngtrees+1); i++) {
    ppLikMat[i] = (double*)malloc((nspecies+1)*sizeof(double));
  }


  ppTipMat = (double**)malloc((ngtrees+1)*sizeof(double*));
  for (i=0; i<(ngtrees+1); i++) {
    ppTipMat[i] = (double*)malloc((nspecies+1)*sizeof(double));
  }


  for (i=0; i<2; i++)
    {
      ppTwoRowS_temp[i] = (int*)malloc((2*nspecies)*sizeof(int));
      if (ppTwoRowS_temp[i]==NULL)
	{
	  printf("Can't memalloc ppTwoRowS_temp[%d]\n",i);
	  return 0;
	}
    }


  for (i=0; i<2; i++)
    {
      ppTwoRow_temp[i] = (int*)malloc((2*ntaxa)*sizeof(int));
      if (ppTwoRow_temp[i]==NULL)
	{
	  printf("Can't memalloc ppTwoRow_temp[%d]\n",i);
	  return 0;
	}
    }

  pppTR = (int***)malloc((2*ngtrees+1)*sizeof(int**));
  if (pppTR==NULL) {
    printf("Can't memalloc pppTR\n");
    return 0;
  }

  for (i=0; i<2*ngtrees+1; i++) {
    pppTR[i] = (int**)malloc(3*sizeof(int*));
    for (j=0; j<3; j++) {
      pppTR[i][j] = (int*)malloc((ntaxa+1)*sizeof(int));
    }
  }

  pppTRS = (int***)malloc((2*num_saved_trees+1)*sizeof(int**));
  if (pppTRS==NULL) {
    printf("Can't memalloc pppTRS\n");
    return 0;
  }
  for (i=0; i<2*num_saved_trees+1; i++) {
    pppTRS[i] = (int**)malloc(3*sizeof(int*));
    for (j=0; j<3; j++) {
      pppTRS[i][j] = (int*)malloc((nspecies+1)*sizeof(int));
    }
  }

  parents = (int*)malloc((2*ntaxa+1)*sizeof(int));
  if (parents==NULL) 
    {
      printf("Can't memalloc parents\n");
      return 0;
    }
  
  ChangeNodeLength = (double*)malloc((2*ntaxa+1)*sizeof(double));
  if (ChangeNodeLength==NULL)
    {
      printf("Can't memalloc ChangeNodeLength\n");
      return 0;
    }

  filled_ind = (int*)malloc((2*ntaxa)*sizeof(int));
  
  filled_indS = (int*)malloc((2*nspecies)*sizeof(int));

  ppLengthMat = (double**)malloc((2*ntaxa+1)*sizeof(double*));
  if (ppLengthMat==NULL)
    {
      printf("Can't memalloc ppLengthMat\n");
      return 0;
    }
  
  for (i=0; i<2*ntaxa+1; i++)
    {
      ppLengthMat[i] = (double*)malloc((2*ntaxa+1)*sizeof(double));
      if (ppLengthMat[i]==NULL)
        {
          printf("Can't memalloc ppLengthMat[%d]\n",i);
          return 0;
        }
    }

  ppLengthMatS = (double**)malloc((2*nspecies+1)*sizeof(double*));
  if (ppLengthMatS==NULL)
    {
      printf("Can't memalloc ppLengthMatS\n");
      return 0;
    }

  for (i=0; i<2*nspecies+1; i++)
    {
      ppLengthMatS[i] = (double*)malloc((2*nspecies+1)*sizeof(double));
      if (ppLengthMatS[i]==NULL)
        {
          printf("Can't memalloc ppLengthMatS[%d]\n",i);
          return 0;
        }
    }


  ppSpecCon = (double**)malloc((nspecies+1)*sizeof(double*));
  if (ppSpecCon==NULL)
    {
      printf("Can't memalloc ppSpecCon\n");
      return 0;
    }
  
  for (i=0; i<2*nspecies+1; i++)
    {
      ppSpecCon[i] = (double*)malloc((nspecies+1)*sizeof(double));
      if (ppSpecCon[i]==NULL)
        {
          printf("Can't memalloc ppSpecCon[%d]\n",i);
          return 0;
        }
    }


  ppSpecConUsed = (int**)malloc((nspecies+1)*sizeof(int*));
  if (ppSpecConUsed==NULL)
    {
      printf("Can't memalloc ppSpecConUsed\n");
      return 0;
    }

  for (i=0; i<2*nspecies+1; i++)
    {
      ppSpecConUsed[i] = (int*)malloc((nspecies+1)*sizeof(int));
      if (ppSpecConUsed[i]==NULL)
        {
          printf("Can't memalloc ppSpecConUsed[%d]\n",i);
          return 0;
        }
    }
  

  CoalVec = (int*)malloc((2*nspecies+1)*sizeof(int));
  if (CoalVec==NULL)
    {
      printf("Can't memalloc CoalVec\n");
      return 0;
    }

  CoalVec2 = (int*)malloc((2*ntaxa+1)*sizeof(int));
  if (CoalVec2==NULL)
    {
      printf("Can't memalloc CoalVec\n");
      return 0;
    }

  CoalVecL = (int*)malloc((2*nspecies+1)*sizeof(int));
  if (CoalVecL==NULL)
    {
      printf("Can't memalloc CoalVecL\n");
      return 0;
    }

  CoalVecR = (int*)malloc((2*nspecies+1)*sizeof(int));
  if (CoalVecR==NULL)
    {
      printf("Can't memalloc CoalVecR\n");
      return 0;
    }


  recalced = (int*)malloc((2*nspecies)*sizeof(int));
  if (recalced==NULL)
    {
      printf("Can't memalloc recalced\n");
      return 0;
    }

  ppMatrix = (int**)malloc(nspecies*sizeof(int*));
  if (ppMatrix==NULL)
    {
      printf("Can't memalloc ppMatrix\n");
      return 0;
    }

  for (i=0; i<nspecies; i++)
    {
      ppMatrix[i] = (int*)malloc(nspecies*sizeof(int));
      if (ppMatrix[i]==NULL)
        {
          printf("Can't memalloc ppMatrix[%d]\n",i);
          return 0;
        }
    }

  SpecCount = (int*)malloc((nspecies+2)*sizeof(int));

  SpecVec = (int**)malloc((nspecies+2)*sizeof(int*));
  for (i=0; i<(nspecies+2); i++) {
    SpecVec[i] = (int*)malloc((ntaxa+1)*sizeof(int));
  }

  return 1;

}


/* function to set seeds from /dev/random is possible */

long int random_seed()
{

 unsigned long int seed;
 struct timeval tv;
 FILE *devrandom;

 if ((devrandom = fopen("/dev/random","r")) == NULL) {
   gettimeofday(&tv,0);
   seed = tv.tv_sec + tv.tv_usec;
   //printf("  Got seed %u from gettimeofday()\n",seed);
 } else {
   fread(&seed,sizeof(seed),1,devrandom);
   //printf("   Got seed %u from /dev/random\n",seed);
   fclose(devrandom);
 }

 return((long int)(seed/10000));

}

int main() {
  
  FILE *gtf, *resi, *set, *sinfo, *stf;
  int i, j, k, check;
  int snum, nlin, label;
  int tax_counter, lin_counter;
  long int seed1, seed2;
  double total_time=0.0, grate;
  float pbound_un, pburnin;
  char *opt_name[18];
  struct Node n;
  FILE *treefile;

  //printf("\n\n>>> STEM: Species Tree Estimation using Maximum likelihood <<<\n\n");


  /**** Read info from settings file ****/

  set = fopen("settings","r");
  fscanf(set,"%*[^\n]"); 
  fscanf(set,"%18s %d",&opt_name,&search);
  fscanf(set,"%*[^\n]"); 
  fscanf(set,"%18s %f",&opt_name,&theta);
  fscanf(set,"%18s %d",&opt_name,&num_saved_trees);
  fscanf(set,"%18s %d",&opt_name,&seed1);
  fscanf(set,"%18s %d",&opt_name,&seed2);
  fscanf(set,"%18s %d",&opt_name,&verbose);

  fscanf(set,"%18s %18s %18s %18s %18s %18s %18s %18s",&opt_name,&opt_name,&opt_name,&opt_name,&opt_name,&opt_name,&opt_name,&opt_name);
  fscanf(set,"%18s %f",&opt_name,&beta);
  fscanf(set,"%18s %d",&opt_name,&burnin);
  fscanf(set,"%18s %d",&opt_name,&bound_total_iter);
  fscanf(set,"%18s %f",&opt_name,&pbound_un);
  fscanf(set,"%18s %f",&opt_name,&pburnin);
  
  fscanf(set,"%18s %18s %18s",&opt_name,&opt_name,&opt_name);
  fscanf(set,"%18s %d",&opt_name,&nspecies);
  fscanf(set,"%18s %d",&opt_name,&ntaxa);
  fscanf(set,"%18s %d",&opt_name,&ngtrees);
  printf("Read settings file successfully.\n");
  
  check = MemAlloc();

  taxname = (naym*)malloc(ntaxa*sizeof(naym));
  speciesname = (naym*)malloc(nspecies*sizeof(naym));

  printf("Completed memory allocation.\n");

  tax_counter = 0;
  for (i=0; i<nspecies; i++) {
    lin_counter=0;
    fscanf(set,"%s %s",speciesname[i],taxname[tax_counter]);
    lin_counter++;
    //printf("species name is %s and taxon name is %s\n",speciesname[i],taxname[tax_counter]);
    while (strcmp(taxname[tax_counter],";")!=0) {
      tax_counter++;
      fscanf(set, "%s",taxname[tax_counter]);
      //printf("Next taxon name is %s\n",taxname[tax_counter]);
      lin_counter++;
    }
    SpecCount[i+1] = lin_counter-1;
    for (j=0; j<SpecCount[i+1]; j++) SpecVec[i+1][j] = tax_counter - lin_counter + j + 2;
    if (i==0) fscanf(set,"%18s %18s %18s %18s %18s %18s %18s %18s",&opt_name,&opt_name,&opt_name,&opt_name,&opt_name,&opt_name,&opt_name,&opt_name);
  }
  if (tax_counter != ntaxa) {
    printf("Error in reading settings file:");
    printf("The number of specified taxa was %d, but %d taxon names were read\n",ntaxa,tax_counter);
    exit(1);
  }

  if (search==2){
    printf("  Setting random number seeds ....\n"); 
    if (seed1==0 && seed2==0) { /* set seeds from system */
      seed1 = random_seed();
      seed2 = random_seed();
    }
  }
  
  CreateList();
  bound_last_un = cbound*ntaxa;
  cbound_un = (int)ceil(log(pbound_un)/(ntaxa*log(1.0-(1.0/ntaxa))));
  bound_un = cbound_un*3*ntaxa;
  extra_burnin = burnin + (int)ceil(log(pburnin)/(ntaxa*log(1.0-(1.0/ntaxa))))*ntaxa;

  printf("\nDone reading species and taxon info from settings file:\n\n");
  for (i=1; i<nspecies+1; i++) {
    printf("  For species %d, there are %d taxa: ",i,SpecCount[i]);
    printf("  For species %s, there are %d taxa: ",speciesname[i-1],SpecCount[i]);
    for (j=0; j<SpecCount[i]; j++) printf(" %s",taxname[SpecVec[i][j]-1]);
    printf("\n");
  }
  printf("\n\n%d gene trees will be read from file genetrees.tre:\n\n",ngtrees);

  fclose(set);

  /**** End read settings from file ****/


  gtf=fopen("genetrees.tre","r");
  resi=fopen("results","w"); 
  
  /* Read in gene trees and branch lengths */
  for (i=0; i<ngtrees; i++) {

    //printf("  Reading gene tree %d\n",i);

    /* clear stored trees */
    for (k=0; k<ntaxa-1; k++) {
      ppTwoRow_temp[0][k]=0;
      ppTwoRow_temp[1][k]=0;
      filled_ind[k]=0;
    }
    //printf("Filled ppTwoRow array.\n");
    for (k=0; k<2*ntaxa; k++) {
      TimeVec_temp[k] = 0.0;
      for (j=0; j<2*ntaxa; j++) ppLengthMat[k][j] = 0.0;
    }
    //printf("Filled TimeVec array.\n");

    /* now read next tree and prepare for storage */
    grate = ReadTree(gtf);
    //printf("Read tree %d\n", i);

    /*
    printf("ppLengthMatrix looks like this:\n");

    int z;
    for(z = 0; z < (2*ntaxa+1); z ++)
    {
        int y;
        for(y = 0; y < (2*ntaxa+1); y ++)
            printf("[%6f]", ppLengthMat[z][y]);
        printf("\n");
    }
    */

    total_time = FindTotalTime();
    //printf("Found the totaltime.");
    CalcTimeVec(total_time,grate);
    //printf("Calculated time vector.\n");
    /* store tree permanently */
    for (j=0; j<ntaxa-1; j++) {
      pppTR[i][0][j] = ppTwoRow_temp[0][j];
      pppTR[i][1][j] = ppTwoRow_temp[1][j];
    }

    for (j=0; j<2*ntaxa; j++) ppTimeVec[i][j] = TimeVec_temp[j];

  }
  fclose(gtf);
  printf("Closed gtf.\n");

  if (verbose == 1) {
    /* Check - see that pppTR is correct */
    for (i=0; i<ngtrees; i++) {
      printf("Tree %d:\n",i);
      for (j=0; j<ntaxa-1; j++) {
	printf("%d %d\n",pppTR[i][0][j],pppTR[i][1][j]);
      }
      printf("\n");
    }
    /* Check - see that ppTimeVec is ok */
    for (i=0; i<ngtrees; i++) {
      printf("Tree %d has vector:\n",i);
      for (j=0; j<2*ntaxa; j++) {
	printf("%d %f\n",j,ppTimeVec[i][j]);
      }
      printf("\n");
    }
  }

  /* Next get info on species memberships for all taxa */
  //  sinfo = fopen("species.info","r");
  //for (i=0; i<nspecies; i++) {
  //  fscanf(sinfo,"%d %d",&snum,&nlin);
  //  SpecCount[snum]=nlin;
  //  for (j=0; j<nlin; j++) {
  //    fscanf(sinfo,"%d",&label);
  //    SpecVec[snum][j] = label;
  //  }
  //  SpecVec[snum][nlin]=0;
  //}

  /* Check */
  //printf("\n");
  //for (i=1; i<nspecies+1; i++) {
  //  printf("  For species %d, there are %d taxa: ",i,SpecCount[i]);
  //  for (j=0; j<SpecCount[i]; j++) printf("%d ",SpecVec[i][j]);
  //  printf("\n");
  // }
  //printf("\n");

  /* Get matrix of pairwise limits on species divergence times */
  GetSpecCon();
  //printf("\n");

  //printf("  Please note: If there are many ties and/or many zeroes or near zeroes\n");
  //printf("  in the matrix above, many species trees will be tied in likelihood score.\n");
  //printf("  You should run a simulated annealing search to collect information\n");
  //printf("  on the set of trees with high likelihood.\n\n");

  if (search == 1) { /* Get ML tree */

    build_ML_tree();
    previous_lik = GetLik();
    printf("\n\n>>>The ML species tree is ");
    write_species_tree(nspecies+1,nspecies+1);
    printf(" with likelihood %f\n\n",previous_lik);

    /***  put ML tree into linked list in order to write output to file  ***/ 
    make_indmat();
    n.iTree = make_matnum(ppMatrix);
    currenttree = AddNodeAscend(&n,previous_lik,resi);
    treefile = fopen("treefile.phy","w");
    write_tree_root_saved(nspecies+1,nspecies+1,currenttree,treefile);
    fprintf(treefile,";\n");
    fprintf(treefile, "%f\n\n", previous_lik);
    
    fclose(treefile);

    fprintf(resi,"\n\nAll results have been written to standard output.  In addition, the ML tree has been written to the file 'treefile.phy'.\n\n");
    printf("\nInformation about trees has been written to the file 'treefile.phy'.\n\n  Good-bye!\n\n");
  }

  else if (search == 0) { /* read in user tree */

    printf("  Reading species tree from file ......");
    fflush(0);
    for (i=0; i<2*nspecies; i++) TimeVecS[i] = 0.0;
    stf = fopen("speciestree.tre","r");
    ReadSpeciesTree(stf);
    printf(" done.");
    fflush(0);
    if (verbose == 1 ) for (i=0; i<nspecies; i++) printf("%d %d\n",ppTwoRowS[0][i],ppTwoRowS[1][i]);
    GetMLLengths();
    previous_lik = GetLik();
    printf("\n\n>>>The likelihood is %f",previous_lik);
    printf(" for tree: ");
    write_species_tree(nspecies+1,nspecies+1);
    if (verbose == 1) {
      for (i=0; i<nspecies-1; i++) printf("%d %d %f\n",ppTwoRowS[0][i],ppTwoRowS[1][i],TimeVecS[i+nspecies+1]);
    }
    printf(";\n\nInformation about trees has been written to the file 'treefile.phy'.\n\n  Good-bye!\n\n");
  }

  else if (search==2) { /* do annealing */

    build_random_tree(seed1,seed2);
    GetMLLengths();
    previous_lik = lik = GetLik();
    if (verbose == 1) printf("The maximum likelihood of the initial random tree is %f\n",previous_lik);

    /***  add initial tree to linked list of saved trees  ***/ 
    make_indmat();
    n.iTree = make_matnum(ppMatrix);
    currenttree = AddNodeAscend(&n,previous_lik,resi);
    
    anneal(resi,seed1,seed2);

    fprintf(resi,">> There were %d topology changes accepted.\n\n\n",no_topo_changes);

    PrintTiedTrees(resi);

    printf("  Information about all trees found in the search has been written to the files 'results' and 'treefile.phy'.\n\n  Good-bye!\n\n");
    
  }

  else {

    printf("  The setting for 'search?' in the settings file must be either\n\n  0 = compute likelihood for species tree in file 'speciestree.tre'\n     OR\n  1 = return maximum likelihood tree and branch lengths only\n      OR\n  2 = search for high likelihood species trees.\n\nPlease change this setting and re-run the program.\n\n\n");

  }  

  //fclose(stf);
  fclose(resi);
  //printf("Returning successfully\n");
  //printf("%d",-1);
  return 0;  
}



