#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#define NR_END 1
#define FREE_ARG char*


/*****  UTILITY FILES FROM NUMERICAL RECIPES  *****/


void nrerror(char error_text[]) {  /* NR standard error handler */
  fprintf(stderr, "Numerical Recipes run-time error ...\n");
  fprintf(stderr, "%s\n",error_text);
  fprintf(stderr, "... now exiting to system ...\n");
  exit(1);
}

double *vector(long nl, long nh) {
  
  double *v;

  v = (double *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(double)));
  if (!v) nrerror("allocation failure in vect()");
  return v-nl+NR_END;
}

double **matrix(long nrl, long nrh, long ncl, long nch) {
  
  long i, nrow = nrh-nrl+1, ncol = nch-ncl+1;
  double **m;
  
  m = (double **) malloc((size_t)((nrow+NR_END)*sizeof(double*)));
  if (!m) nrerror("allocation failure 1 in matrix()");
  m += NR_END;
  m -= ncl;

  m[nrl] = (double *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(double)));
  if (!m[nrl]) nrerror("allocation failure 2 in matrix()");
  m[nrl] += NR_END;
  m[nrl] -= ncl;

  for (i=nrl+1; i<=nrh; i++) m[i] = m[i-1] + ncol;

  return m;

}

void free_vector(double *v, long nl, long nh) {
  
  free((FREE_ARG) (v+nl-NR_END));

}

void free_matrix(double **m, long nrl, long nrh, long ncl, long nch) {

  free((FREE_ARG) (m[nrl]+ncl-NR_END));
  free((FREE_ARG) (m+nrl-NR_END));

}
