#include <stdio.h>
#include <math.h>
#include "main.h"
#include "linpack.c"
#include "ranlib.h"
#include "ranlib.c"
#include "com.c"


/**********************************************/
/***  Functions involved in tree building   ***/
/**********************************************/

void build_random_tree(int seedj, int seedk) {

  int i;
  int count = 0, found_one = 0;
  int curr_length;
  int *ppNodes[2];
  int node1, node2;
  int new_counter;

  /* allocate memory */
  
  for (i=0; i<2; i++)
    {
      ppNodes[i] = (int*)malloc(nspecies*sizeof(int));
      if (ppNodes[i]==NULL)
	{
	  printf("Can't memalloc ppNodes[%d]\n",i);
	}
    }
  
  /* done memory allocation */

  curr_length=nspecies;
  new_counter=nspecies+2;

  for (i=0; i<curr_length; i++) {    
    ppNodes[0][i]=i+1;
    ppNodes[1][i]=0;    
  }
  
  /*  construct the random tree */

  while (count<nspecies-2) {
    
    /* randomly pick two nodes to join by generating random */
    /* numbers from a uniform distribution                  */
    
    node1 = (int)(ranf()*(curr_length));
    node2 = (int)(ranf()*(curr_length));

     while (node1==node2) {

      node2 = (int)(ranf()*(curr_length));
       
    }
    
    ppTwoRowS[0][new_counter-(nspecies+1)]=ppNodes[0][node1];
    ppTwoRowS[1][new_counter-(nspecies+1)]=ppNodes[0][node2];
    
    ppNodes[1][node1]=1;
    ppNodes[1][node2]=1; 
    
    /* shift */

    found_one = 0;

    for (i=0; i<curr_length; i++ ) {
      
      if (ppNodes[1][i]==1 && found_one==0) {
	
	ppNodes[0][i]=new_counter;
	ppNodes[1][i]=0;
	found_one=1;
	
      }
      
      if (ppNodes[1][i]==1 && found_one==1) found_one=2;
      
      if (found_one==2) {
	
	ppNodes[0][i]=ppNodes[0][i+1];
	
      }
      
    }
    
    curr_length = curr_length-1;
    
    for (i=0; i<curr_length; i++) {
      
      ppNodes[1][i]=0;
      
    }
    
    /* done shift */
    
    count++;
    new_counter++;
    
  }
  
  ppTwoRowS[0][0]=ppNodes[0][0];
  ppTwoRowS[1][0]=ppNodes[0][1];


}

/**** Function used in constructing ML tree ****/

void FindMin(int *node1, int *node2, double *min_time) {

  int i, j;
  double tbound, current_min;

  tbound = *min_time;
  current_min = 1000.0;
  for (i=1; i<nspecies+1; i++) {
    for (j=i+1; j<nspecies+1; j++) {
      if (ppSpecCon[i][j] < current_min && ppSpecCon[i][j] >= tbound && ppSpecConUsed[i][j]==0) {
	current_min = ppSpecCon[i][j];
	*node1 = i;
	*node2 = j;
      }
    }
  }

  ppSpecConUsed[*node1][*node2] = 1;
  *min_time = current_min;
  return;

}

int FindParentNode(int target) { 

  int i, j, parent=0;

  for (j=0; j<2; j++) {
    for (i=0; i<nspecies; i++) {
      if (ppTwoRowS[j][i]==target) {
        parent = i+nspecies+1;
        break;
      }
    }
  }

  /*  if (parent==0) { printf("Could not find parent for node %d\n",target); }*/
  return(parent);
}



/**** Function to construct the ML tree ****/

void build_ML_tree() {

  int i, j, k, l;
  int count = 0, found_one = 0, onode, parent, parent1, parent2, check=0, check2=0;
  int curr_length;
  int *ppNodes[2];
  int node1=0, node2=0;
  int new_counter;
  double div_min=0.0;

  int tempcount=0;

  /* allocate memory */
  
  for (i=0; i<2; i++)
    {
      ppNodes[i] = (int*)malloc(nspecies*sizeof(int));
      if (ppNodes[i]==NULL) printf("Can't memalloc ppNodes[%d]\n",i);
    }
  
  /* done memory allocation */

  for (i=1; i<nspecies+1; i++) {
    for (j=i+1; j<nspecies+1; j++) ppSpecConUsed[i][j] = 0;
  }

  curr_length=nspecies;
  new_counter=nspecies+2;

  for (i=0; i<curr_length; i++) {    
    ppNodes[0][i]=i+1;
    ppNodes[1][i]=0;    
  }
  
  for (i=0; i<nspecies; i++) ppTwoRowS[0][i] = ppTwoRowS[1][i] = 0;

  /*  construct the tree */

  while (count<nspecies-2) {
   
    FindMin(&node1,&node2,&div_min);

    /* check to see if both, one, or neither of the returned nodes are still available */

    check = 0;
    for (k=0; k<curr_length; k++) {
      if (ppNodes[0][k]==node1) check++;
      if (ppNodes[0][k]==node2) check++;
    }

    if (check==2) {/* both are still available, so join them */
      
      for (k=0; k<curr_length; k++) if (ppNodes[0][k] == node1) break;
      ppTwoRowS[0][new_counter-(nspecies+1)]=ppNodes[0][k];
      ppNodes[1][k]=1;
      for (k=0; k<curr_length; k++) if (ppNodes[0][k] == node2) break;
      ppTwoRowS[1][new_counter-(nspecies+1)]=ppNodes[0][k];
      ppNodes[1][k]=1;

      TimeVecS[new_counter] = div_min;

    }

    else if (check==0) {  /* neither is available - see if they share a common ancestor */

      check2=0;
      parent1 = FindParentNode(node1);
      parent2 = FindParentNode(node2);
      
      while (parent1 != nspecies+1 && check2 != 1) {
	while (parent2 != nspecies+1 && check2 != 1) {
	  if (parent1 == parent2) {
	    check2 = 1;
	    break;
	  }
	  parent2 = FindParentNode(parent2);
	}
	if (check2 == 1) break;
	else {
	  parent1 = FindParentNode(parent1);
	  parent2 = FindParentNode(node2);
	}
      }

      if (parent1==parent2 && parent1!=0 && (parent1<nspecies+1 || parent2!=0)) { } /* they already share an MRCA - find next smallest time */
      else {  /* they both have coalesced, but not with each other - join their clades */

	onode = node1;
	while (onode!=0) {
	  parent=onode;
	  onode = FindParentNode(parent);
	}
	l=0;
	while (l<curr_length) {
	  if (ppNodes[0][l] == parent) break;
	  l++;
	}

	ppTwoRowS[0][new_counter-(nspecies+1)]  = ppNodes[0][l];
	TimeVecS[new_counter] = div_min;
	ppNodes[1][l] = 1;
	
	onode = node2;
        while (onode!=0) {
          parent=onode;
          onode = FindParentNode(parent);
        }
        l=0;
        while (l<curr_length) {
          if (ppNodes[0][l] == parent) break;
          l++;
        }
        ppTwoRowS[1][new_counter-(nspecies+1)]  = ppNodes[0][l];
        ppNodes[1][l] = 1;
	
      }

    }


    else { /* one is available, so add it to tree */

      k=0;
      while (k<curr_length) {
	if (ppNodes[0][k]==node1) { /* node1 is available */
	  ppTwoRowS[0][new_counter-(nspecies+1)] = ppNodes[0][k];
	  onode = node2;
	  while (onode!=0) {
	    parent=onode;
	    onode = FindParentNode(parent);
	  }
	  l=0;
	  while (l<curr_length) {
	    if (ppNodes[0][l] == parent) break;
	    l++;
	  }
	  ppTwoRowS[1][new_counter-(nspecies+1)] = ppNodes[0][l];
	  TimeVecS[new_counter] = div_min;
	  ppNodes[1][k] = 1;
	  ppNodes[1][l] = 1;
	  break;
	}
	else if (ppNodes[0][k]==node2) { /* node2 is available */
	  ppTwoRowS[0][new_counter-(nspecies+1)] = ppNodes[0][k];
	  onode = node1;
	  while (onode!=0) {
	    parent=onode;
	    onode = FindParentNode(parent);
	  }
	  l=0;
	  while (l<curr_length) {
	    if (ppNodes[0][l] == parent) break;
	    l++;
	  }
	  ppTwoRowS[1][new_counter-(nspecies+1)] = ppNodes[0][l];
	  TimeVecS[new_counter] = div_min;
	  ppNodes[1][k] = 1;
	  ppNodes[1][l] = 1;
	  break;
	}

	k++;
      }

    }

    /* shift */

    found_one = 0;
    
    for (i=0; i<curr_length; i++ ) {
      
      if (ppNodes[1][i]==1 && found_one==0) {
	
	ppNodes[0][i]=new_counter;
	ppNodes[1][i]=0;
	found_one=1;
	
      }
      
      if (ppNodes[1][i]==1 && found_one==1) found_one=2;
      
      if (found_one==2) {
	
	ppNodes[0][i]=ppNodes[0][i+1];
	
      }
      
    }
    
    if (found_one!=0) {

      curr_length = curr_length-1;      
      for (i=0; i<curr_length; i++) ppNodes[1][i]=0;
      
    }
    
    /* done shift */
    
    if (found_one==1 || found_one==2) {
      count++;
      new_counter++;
    }

    /*for (i=0; i<nspecies-1; i++) printf("%d %d %f\n",ppTwoRowS[0][i],ppTwoRowS[1][i],TimeVecS[i+nspecies+1]); */
    
  }


  parent1 = parent2 = 0;
  while (parent1==parent2) {
    FindMin(&node1,&node2,&div_min);
    parent1 = node1;
    while (parent1 != ppNodes[0][0] && parent1 != ppNodes[0][1]) parent1=FindParentNode(parent1);
    parent2 = node2;
    while (parent2 != ppNodes[0][0] && parent2 != ppNodes[0][1]) parent2=FindParentNode(parent2);
  }

  ppTwoRowS[0][0]=ppNodes[0][0];
  ppTwoRowS[1][0]=ppNodes[0][1];
  TimeVecS[nspecies+1] = div_min;

  if (verbose == 1) for (i=0; i<nspecies-1; i++) printf("%d %d %f\n",ppTwoRowS[0][i],ppTwoRowS[1][i],TimeVecS[i+nspecies+1]);

}



/*** The local rearrangement procedure ***/

void trbldg(FILE *res) {

  int jj;
  int i, index=0, target, parent=0, left_child, right_child, sib, side_target;
  double testlik;
  float ran_num2;
  struct Node n;

  /* Put current tree in ppTwoRow_temp */
  for (i=0; i<nspecies-1; i++) {
    ppTwoRowS_temp[0][i] = ppTwoRowS[0][i];
    ppTwoRowS_temp[1][i] = ppTwoRowS[1][i];
  }
  for (i=1; i<2*nspecies; i++) TimeVecS_temp[i] = TimeVecS[i];


  /* Randomly choose an internal node as the target node */  
  target = (int)(ranf()*(nspecies-1))+nspecies+1;

  /* Identify parent, right_child, left_child, and sib for */
  /* the target.  Also let the variable side_target        */
  /* be zero if the target is on the parent's left side    */
  /* and one if the target is on the parent's right side.  */
  
  parent = FindParent(target);
  left_child = ppTwoRowS[0][target-(nspecies+1)];
  right_child = ppTwoRowS[1][target-(nspecies+1)];
  if (ppTwoRowS[0][parent-(nspecies+1)] == target) {    
    sib = ppTwoRowS[1][parent-(nspecies+1)];
    side_target=0;
  }  
  else { 
    sib = ppTwoRowS[0][parent-(nspecies+1)];
    side_target = 1;
 }

  /* Generate 1, 2, or 3 at random to decide which of the three  */
  /* possible local rearrangements will be made.  1 = no change, */
  /* 2 = move left_child to branch connecting sib and parent,    */
  /* 3  = move right_child to branch connecting sib and parent   */

  /* only do this if target is not equal to the root */
  
  if (target > nspecies+1) {
    index=(int)(ranf()*3)+1;
    if (index>3 || index<1) {
      printf("There was an index generating problem in file trbldg.c at line=%d",__LINE__);
      printf("\n\n");
      exit(1);
    }
    
    if (index==2) {
      
      ppTwoRowS[side_target][parent-(nspecies+1)]=right_child;
      ppTwoRowS[(side_target+1)%2][parent-(nspecies+1)]=target;
      ppTwoRowS[side_target][target-(nspecies+1)]=left_child;
      ppTwoRowS[(side_target+1)%2][target-(nspecies+1)]=sib;
      
    }
    
    else { 
      if (index==3){
	
	ppTwoRowS[side_target][parent-(nspecies+1)]=left_child;
	ppTwoRowS[(side_target+1)%2][parent-(nspecies+1)]=target;
	ppTwoRowS[side_target][target-(nspecies+1)]=right_child;
	ppTwoRowS[(side_target+1)%2][target-(nspecies+1)]=sib;
      }
    }
  }

  if (verbose == 1) {
    for (jj=0; jj<nspecies-1; jj++) printf("%d %d\n",ppTwoRowS[0][jj],ppTwoRowS[1][jj]);
    fflush(0);
  }

  /*Get ML branch lengths */
  
  GetMLLengths();
  testlik = GetLik();

  /* determine whether to accept change */
  /* if new likelihood is larger, always accept */
  /* if smaller, accept probabilistically */

  if (testlik < previous_lik) {

    ran_num2 = ranf();
    if (ran_num2<=exp((testlik-previous_lik)/ci)) { /*accept - nothing to be done*/
      previous_lik=testlik;

      no_accept++;
      if (index==2 || index==3) {
	no_topo_changes++;
	if (num_saved_trees==0) no_last_un=1;
      
	/* add tree info to linked list if appropriate */
	if (done_burnin==1 && num_saved_trees>0) {
        
	  make_indmat();
	  n.iTree = make_matnum(ppMatrix);
	  currenttree = AddNodeAscend(&n,previous_lik,res);

	}
      }
    
      else {
	currenttree->times_hit++;
	if (currenttree->times_hit<bound_un) no_last_un=1;
      }
    }
    
    else {  /* reject - put things back and fix ppLikMat */

      for (i=0; i<nspecies; i++) {
	ppTwoRowS[0][i] = ppTwoRowS_temp[0][i];
	ppTwoRowS[1][i] = ppTwoRowS_temp[1][i];
      }
      for (i=1; i<2*nspecies; i++) TimeVecS[i] = TimeVecS_temp[i];
      GetMLLengths();
      previous_lik = GetLik(); /* this is lazy - just fix ppLikMat for changed nodes */
                               /* as we do in file branch.c                          */
      currenttree->times_hit++;
      no_last_un++;

    }

  }

  else {
    previous_lik = testlik; 

    no_accept++;
    if (index==2 || index==3) {

      no_topo_changes++;
      
      if (num_saved_trees==0) no_last_un=1;

      /* add tree info to linked list if appropriate */
      if (done_burnin==1 && num_saved_trees>0) {
        
        make_indmat();
        n.iTree = make_matnum(ppMatrix);
        currenttree = AddNodeAscend(&n,previous_lik,res);

      }
      
    }

  }  
}
